#!/usr/bin/perl
#
# A generic HTML interface to a DBM Table
# Stores arrays in a dbm database
#
# By Evan Jones <ejones@uwaterloo.ca>
# http://www.eng.uwaterloo.ca/~ejones/
#
# Released under the BSD Licence:
# - Essentially, I hold copyright but you can do whatever you want with this,
#	as long as you acknowledge my copyright.
# http://www.opensource.org/licenses/bsd-license.html
#
# TODO:
# - Make it easy to install (a configuration file?)
# - Move the configuration into the DBM file: More generic
#	- CSS, title, etc
# - Add support for basic HTTP authentication into the script itself
# - Add table sizing
# - File locking?
#
# Feature Enhancements:
# - Create an administrator interface which can configure database files
# - Use a *real* database
# - Defined data types (and therefore, input validation)
# - Processing of data before it is displayed eg. Transform "1234567890" -> "(123) 456-7890"
# - Links between tables eg. "Location = 5" -> "Location = Calgary, Alberta, Canada"

use strict;

my $stylesheet = "/~ejones/database.css";
my $title = "Database";

use Fcntl;
use FindBin;
use lib "$FindBin::Bin";
	   
use DBMTable;

# Maps modes to functions
my %modes = (
	list => \&list,
	edit => \&entryform,
	confirm => \&confirm,
	del => \&del,
	save => \&save,
	create => \&create,
	);

# The tied table variable
my %info;

# Get the script parameters
my %param = getcgivars();

$title .= ": $param{filename}" if ( exists $param{filename} );

# \r\n is specified in the HTTP specification
print( "Content-type: text/html\r\n\r\n" );
print <<EOF;
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html><head>
<title>$title</title>
<link rel="stylesheet" href="$stylesheet" type="text/css" />
</head><body>
EOF

#~ if ( exists $param{debug} )
#~ {
	#~ debug( \%param );
#~ }
if ( exists $param{filename} && $param{filename} ne '' )
{
	if ( $param{filename} !~ /^[A-Za-z0-9\.]+$/ )
	{
		error( "The file name can only contain letters, numbers and periods." );
	}
	else
	{
		if ( tie( %info, 'DBMTable', $param{filename} ) )
		{
			if ( $param{mode} eq 'create' )
			{
				error( "The specified filename, <code>$param{filename}</code>, already exists." )
			}
			else
			{
				if ( exists $param{mode} and exists $modes{ $param{mode} } )
				{
					&{$modes{ $param{mode} }}( \%info, \%param );
				}
				else
				{
					list( \%info, \%param );
				}
			}
		}
		else
		{
			if ( $param{mode} ne 'create' )
			{
				error( "The specified database, <code>$param{filename}</code>, could not be opened." );
			}
			else
			{
				&{$modes{create}}( \%param );
			}
		}
	}
}
else
{
	error( "No database file was specified" );
}

# Clean up
print( "</body></html>\n\n" );
untie( %info );

exit( 0 );

sub debug( % )
{
	my $params = @_[0];
	while ( my ($key, $value) = each( %$params ) )
	{
		print "<p>$key = $value</p>\n";
	}
	
	print "<p>QUERY_STRING = $ENV{QUERY_STRING}</p>\n";
}

# Prints the array as an HTML table
sub displayarray( $$$ )
{
	my ($table, $sort, $filename) = (@_);
	
	my $list = shift( @$table );
	shift( @$list ); # Remove the first element
	
	print( qq{<table>\n<tr>} );

	my $i = 1;  # Since we remove the key column
	foreach my $heading ( @$list )
	{
		print( "<th>" );
		print( qq{<a href="$ENV{SCRIPT_NAME}?filename=$filename&amp;mode=list&amp;$filename&amp;sort=$i">} )
			if ( ! defined $sort || defined $sort and $i != $sort );
		print( $heading );
		print( qq{</a>} ) if ( defined $sort and $i != $sort );
		print( "</th>" );
		++ $i;
	}
	print( "<th>Action</th>" ) if ( @_ > 1 );
	print( "</tr>\n" );

	my $string2 = qq{<a href="$ENV{SCRIPT_NAME}?filename=$filename&amp;mode=};
	my $string1 = $string2 . "edit&amp;key=";
	$string2 .= "confirm&amp;key=";
	foreach my $row ( @$table )
	{
		my $key = shift( @$row );
		print( "<tr><td>", join( "</td><td>", @$row ) );
		print( qq{</td><td class="action">$string1$key">edit</a>&nbsp;|&nbsp;} .
			qq{$string2$key">delete</a>} )
			if ( @_ != 1 );
		print( "</td></tr>\n" );
	}
	print ( "</table>\n" );
}

# Returns an array reference representation of the table
sub buildarray( $$ )
{
	my ($table, $sort) = @_;
	
	my @array = ();
	
	while( my ($key, $value) = each( %$table ) )
	{
		# Skip key zero or named keys
		next if ( $key == 0 );
		push( @array, [ $key, @$value ] );
	}
	
	@array = sort( { $a->[ $sort ] cmp $b->[ $sort ] } @array )
		if ( defined $sort && $sort != -1 );
	
	unshift( @array, [ 'Key', @{$$table{0}} ] );
	
	return \@array;
}

# Outputs the table passed as an argument as an HTML table
sub list( $$ )
{
	my ($table, $param) = @_;
	
	my $sort = ( exists $param->{sort} ) ? $param->{sort} : 1;
	
	displayarray( buildarray( $table, $sort ), $sort, $param->{filename} );
		
	print( qq{<p><a href="$ENV{SCRIPT_NAME}?mode=edit&amp;filename=$param->{filename}">Add a Record</a></p>\n} );
}

sub confirm( $$ )
{
	my ($table, $param) = @_;

	print( "<p>Delete this record?</p>\n" );
	
	my @array = ( [ 'Key', @{$table->{0}} ],
		[ $param->{key}, @{$table->{ $param->{key} }} ] );
	
	displayarray( \@array, undef, $param->{filename} );
	
	print( qq{<p><a href="$ENV{SCRIPT_NAME}?mode=del&amp;filename=$param->{filename}&amp;key=$param->{key}">OK</a>},
		q{&nbsp;|&nbsp;},
		qq{<a href="$ENV{SCRIPT_NAME}?$param->{filename}">Cancel</a></p>\n} );
}

sub del( $$ )
{
	my ($table, $param) = @_;
	
	# Delete the specified key
	if ( exists $param->{key} and exists $table->{ $param->{key} } )
	{
		delete $table->{ $param->{key} };
	}
	
	# Now print the new list
	list( $table, $param );
}

sub save( $$ )
{
	my ($table, $param) = @_;
	
	# Find an available key if no key specified
	# TODO: This can be optimized
	my $key;
	if ( exists $param->{key} )
	{
		$key = $param->{key};
	}
	else
	{
		$key = 1;
		
		while ( exists $table->{$key} )
		{
			++ $key;
		}
	}

	# Save the new data
	my @newdata = ('') x @{$table->{0}};
	for ( my $i = 0; exists $param->{$i}; ++ $i )
	{
		$newdata[$i] = $param{$i};
	}
	$table->{$key} = \@newdata;
	
	# Now print the new list
	list( $table, $param );
}

# Used for adding or editing records
sub entryform( $$ )
{
	my ($table, $param) = @_;
	
	my $data;
	if ( exists $param->{key} )
	{
		$data = $table->{ $param->{key} };
	}
	else
	{
		$data = [ ('') x @{$table->{0}} ];
	}
	
	print( qq{<form method="post" action="$ENV{SCRIPT_NAME}">},
		qq{<input type="hidden" name="mode" value="save" />\n},
		qq{<input type="hidden" name="filename" value="$param->{filename}" />} );
	print( qq{<input type="hidden" name="key" value="$param->{key}"/>\n} )
		if ( exists $param->{key} );
		
	print( qq{<table class="entryform">\n} );
	my $i = 0;
	foreach my $heading ( @{$table->{0}} )
	{
		print( qq{<tr><td class="entryform">$heading</td><td>},
			qq{<input type="text" name="$i" value="$data->[$i]"/>},
			qq{</td></tr>\n} );
		++ $i;
	}
	print( qq{</table>} );
	
	print( qq{<p><input type="submit" value="Save"/>\n} );
	print( "</p></form>\n" );
	print( qq{<p><a href="$ENV{SCRIPT_NAME}?filename=$param->{filename}">List</a></p>\n} );
}

#----------------- start of &getcgivars() module ----------------------
# Stolen from http://www.jmarshall.com/easy/cgi/hello.pl.txt
# - Modified to compile and run with "use strict;" and -w

# Read all CGI vars into an associative array.
# If multiple input fields have the same name, they are concatenated into
#   one array element and delimited with the \0 character (which fails if
#   the input has any \0 characters, very unlikely but conceivably possible).
# Currently only supports Content-Type of application/x-www-form-urlencoded.
sub getcgivars {
    my ($in, $name, $value, %in) ;

    # First, read entire string of CGI vars into $in
    if ( ($ENV{'REQUEST_METHOD'} eq 'GET') ||
         ($ENV{'REQUEST_METHOD'} eq 'HEAD') ) {
        $in= $ENV{'QUERY_STRING'} ;

    } elsif ($ENV{'REQUEST_METHOD'} eq 'POST') {
        if ($ENV{'CONTENT_TYPE'}=~ m#^application/x-www-form-urlencoded$#i) {
            $ENV{'CONTENT_LENGTH'}
                || die("No Content-Length sent with the POST request.") ;
            read(STDIN, $in, $ENV{'CONTENT_LENGTH'}) ;

        } else { 
            die("Unsupported Content-Type: $ENV{'CONTENT_TYPE'}") ;
        }

    } else {
        die("Script was called with unsupported REQUEST_METHOD.") ;
    }
    
    # Resolve and unencode name/value pairs into %in
    if ( defined $in ) {
	foreach (split('&', $in)) {
		s/\+/ /g ;
		($name, $value)= split('=', $_, 2) ;
		$name=~ s/%(..)/chr(hex($1))/ge ;
		$value=~ s/%(..)/chr(hex($1))/ge ;
		$in{$name}.= "\0" if defined($in{$name}) ;  # concatenate multiple vars
		$in{$name}.= $value ;
	}
    }

    return %in;
}

sub create( $ )
{
	my ($param) = @_;
	
	# Create the table
	DBMTable::create( $param->{filename},
		[ split( '\0', $param->{heading} ) ] );
	
	tie( %info, 'DBMTable', $param{filename} ) or die $!;
	list( \%info, $param );
}

sub error( $ )
{
	print( '<h1>Error</h1><p>', $_[0], '</p>' );
}
